import Scope, { ScopeData } from '../Model/Scope';
import { Parser } from "acorn";
import { MessageItem } from "../Model/Message";
import { Node, ESTree } from "../Model/Node";
import { ClosureHandler, BaseClosure, ReturnStringClosure } from '../Model/Closure';
import { MacroTaskList } from '../Model/Task';
export declare let MyParser: typeof Parser;
declare type Getter = () => any;
declare type ECMA_VERSION = 3 | 5 | 6 | 7 | 8 | 9 | 10 | 11 | 2015 | 2016 | 2017 | 2018 | 2019 | 2020;
interface Options {
    ecmaVersion?: ECMA_VERSION;
    /**
     * 沙箱执行超时时间
     * 默认为0，不限制大小
     */
    timeout?: number;
    /**
     * 根作用域，沙箱环境中顶级作用域中访问的this变量，
     * 例如，浏览器环境中的rootContext就是window
     */
    rootContext?: Context | null;
    /**
     * 方法中的根作用域，如果一个没有绑定this变量的方法被调用，
     * 那么这个方法中的this访问是什么呢？通过这个配置指定
     * ```javascript
     * function func(){
     * 		return this;
     * }
     * func()
     * ```
     */
    globalContextInFunction?: any;
    /**
     * 在一个Interpreter实例化之后立即执行
     * 可以用来自定义这个Interpreter实例的任何操作，高阶用法
     * 最常用的，你可以console一下这个实例的globalScope。会加深对这个沙箱环境的理解
     */
    _initEnv?: (this: Interpreter) => void;
}
interface CollectDeclarations {
    [key: string]: undefined | BaseClosure;
}
/**
 * 词法环境
 */
declare type LexContext = {
    [key: string]: {
        kind: 'const' | 'let';
        init: boolean;
    };
};
declare type Context = {
    [prop: string]: any;
    [prop: number]: any;
};
interface GeneratorReflection {
    getOptions(): Readonly<Options>;
    getCurrentScope(): Scope;
    getGlobalScope(): Scope;
    getCurrentContext(): Context;
    getExecStartTime(): number;
}
export declare function internalFunction(reflection: InternalInterpreterReflection, ...params: string[]): (...args: any[]) => any;
export declare class InternalInterpreterReflection {
    protected interpreter: Interpreter;
    constructor(interpreter: Interpreter);
    generator(): GeneratorReflection;
}
declare function internalEval(reflection: InternalInterpreterReflection, code?: string, useGlobalScope?: boolean): any;
export declare class Interpreter extends ClosureHandler {
    static readonly version: string;
    static readonly eval: typeof internalEval;
    static readonly Function: typeof internalFunction;
    static ecmaVersion: ECMA_VERSION;
    static globalContextInFunction: any;
    static global: Context;
    protected value: any;
    protected context: Context | Scope;
    protected globalContext: Context;
    protected source: string;
    protected sourceList: string[];
    protected currentScope: Scope;
    protected globalScope: Scope;
    protected currentContext: Context;
    protected options: Options;
    protected callStack: string[];
    protected taskManager: MacroTaskList;
    protected collectDeclVars: CollectDeclarations;
    protected collectDeclFuncs: CollectDeclarations;
    protected collectDeclLex: LexContext[];
    /**
     * 用于临时存放函数声明的scope，函数声明执行时会用它更新函数体
     */
    protected _functionVarScope: Scope;
    protected isVarDeclMode: boolean;
    protected lastExecNode: Node | null;
    protected isRunning: boolean;
    protected execStartTime: number;
    protected execEndTime: number;
    constructor(context?: Context | Scope, options?: Options);
    /**
     * 初始化全局执行环境
     * @param ctx
     */
    protected initEnvironment(ctx: Context | Scope): void;
    /**
     * 包含块级作用域的语句声明之前执行。创建一个新的hash，记录块作用域内声明的let/const变量
     */
    protected blockDeclareStart(): void;
    /**
     * 包含块级作用域的语句声明之后执行，退栈，获取该块作用域声明的let/const变量
     */
    protected blockDeclareEnd(): LexContext | null;
    getExecStartTime(): number;
    getExecutionTime(): number;
    setExecTimeout(timeout?: number): void;
    getOptions(): Readonly<Options>;
    protected getGlobalScope(): Scope;
    protected getCurrentScope(): Scope;
    protected getCurrentContext(): Context;
    protected isInterruptThrow<T>(err: T): boolean;
    protected createSuperScope(ctx: Context): Scope;
    protected setCurrentContext(ctx: Context): void;
    protected setCurrentScope(scope: Scope): void;
    evaluate(code?: string): any;
    appendCode(code: string): any;
    protected evaluateNode(node: ESTree.Program, source?: string): any;
    protected createErrorMessage(msg: MessageItem, value: string | number, node?: Node | null): string;
    protected createError<T>(message: string, error: {
        new (msg: string): T;
    }): T;
    protected createThrowError<T>(message: string, error: {
        new (msg: string): T;
    }): T;
    protected createInternalThrowError<T extends MessageItem>(msg: T, value: string | number, node?: Node | null): Error;
    protected checkTimeout(): boolean;
    protected getNodePosition(node: (Node & {
        start?: number;
        end?: number;
    }) | null): string;
    protected createClosure(node: Node): BaseClosure;
    protected safeObjectGet(obj: any, key: any, node: Node): any;
    /**
     * @param node
     */
    protected createCallFunctionGetter(node: Node & {
        start?: number;
        end?: number;
    }): () => any;
    protected getIdentifierScope(node: ESTree.Identifier): Scope;
    protected getVariableName(node: ESTree.Pattern): never | string;
    /**
     * 保证变量的访问只能在沙箱内部的。如果一个变量没有找到，会默认返回globalScope。
     * 那么如果globalScope如果不存在name的变量，就throw Error
     * @param data 作用域data
     * @param name Identifier标识符
     * @param node AST: Identifier Node
     */
    protected assertVariable(data: ScopeData, name: string, node: Node): void | never;
    protected createParamNameGetter(node: ESTree.Pattern): ReturnStringClosure;
    protected createObjectKeyGetter(node: ESTree.Expression): Getter;
    protected createMemberKeyGetter(node: ESTree.MemberExpression): Getter;
    /**
     * 如果一个变量作为左值调用，用这个Getter，主要是判断了const
     * 不是作为左值调用，用createObjectGetter即可
     * @param node
     */
    protected createLeftObjectGetter(node: ESTree.Expression | ESTree.Pattern): Getter;
    protected createObjectGetter(node: ESTree.Expression | ESTree.Pattern): Getter;
    protected createNameGetter(node: ESTree.Expression | ESTree.Pattern): Getter;
    protected varDeclaration(name: string): void;
    protected funcDeclaration(name: string, func: () => any): void;
    /**
     * 在当前词法作用域下声明一个词法变量。这里的声明在词法作用域下是提升的，为的是鉴别未初始化之前对词法变量的访问。
     * 如果一个词法变量init:false,说明这个词法变量在该作用域内是存在的，但是还未初始化，访问它就是非法的。
     * @param name
     */
    /**
     * 将实现扫描得到的变量声明和函数申明提前放到scope.data中
     * 注意变量的声明放进去的值只是undefined
     * @param declVars
     * @param declFuncs
     * @param scope
     */
    protected addDeclarationsToScope(declVars: CollectDeclarations, declFuncs: CollectDeclarations, scope: Scope): void;
    protected getScopeValue(name: string, startScope: Scope): any;
    protected getScopeDataFromName(name: string, startScope: Scope): ScopeData;
    /**
     * 从当前scope从下往上找，如果当前scope.type=='block'，
     * 先确保scope.lexDeclared[name]是否为false，
     * 如果为false，说明这个块变量没有初始化，直接报错就行了
     * 如果为true，再去scope.data里面找
     * 如果scope.type=='function'，直接去scope.data里面找。
     * @param name
     * @param startScope
     */
    protected getScopeFromName(name: string, startScope: Scope, constInit?: true): Scope;
    protected entryBlockScope(newScope: Scope): Scope;
    protected setValue(value: any): any;
    getValue(): any;
}
export {};
